#include <windows.h>
#include <process.h>
#include <shlwapi.h>
#include <cstdio>
#include <string>
#include <vector>
#include <cassert>

enum RUNNING_STATE
{
    RUNNING_STATE_NOT_SET,
    RUNNING_STATE_START_MONITORING,
    RUNNING_STATE_STOP,
    RUNNING_STATE_NORMAL
};

class CCriticalSection
{
public:
    CRITICAL_SECTION m_csLock;

    CCriticalSection()
    {
        ::InitializeCriticalSection(&m_csLock);
    }

    virtual ~CCriticalSection()
    {
        ::DeleteCriticalSection(&m_csLock);
    }

    void Lock()
    {
        ::EnterCriticalSection(&m_csLock);
    }

    void Unlock()
    {
        ::LeaveCriticalSection(&m_csLock);
    }
};

class CDirChangeWatcher;

class CDirChangeHandler : public CCriticalSection
{
public:
    CDirChangeHandler();
    virtual ~CDirChangeHandler();

    BOOL UnwatchDir();

    ULONG AddRef();
    ULONG Release();

    virtual BOOL OnChangeNotify(DWORD dwAction, LPCWSTR file1, LPCWSTR file2);

    void SetWatcher(CDirChangeWatcher *pWatcher);

protected:
    LONG m_cRefs;
    CDirChangeWatcher *m_pWatcher;
};

class CDirWatchItem : public CCriticalSection
{
private:
    CDirWatchItem();
    CDirWatchItem(const CDirWatchItem&);
    CDirWatchItem& operator=(const CDirWatchItem&);
    virtual ~CDirWatchItem();

public:
    std::wstring m_strDir;
    HANDLE m_hDir;
    CDirChangeHandler *m_pChangeHandler;
    DWORD m_dwChangeFilter;
    BOOL m_bWatchSubDir;
    HANDLE m_hStartStopEvent;
    RUNNING_STATE m_RunningState;
    DWORD m_dwReadDirError;
    BYTE m_abBuffer[4096];
    OVERLAPPED m_Overlapped;

    CDirWatchItem(LPCWSTR pszDir, HANDLE hDir, CDirChangeHandler *pChangeHandler, DWORD dwChangeFilter, BOOL bWatchSubDir);
    void DeleteSelf(CDirChangeWatcher *pWatcher);

    DWORD StartMonitor(HANDLE hCompPort);
    BOOL StopMonitor(HANDLE hCompPort);
    BOOL SignalShutdown(HANDLE hCompPort);
    BOOL WaitForShutdown();
    void CloseDirectoryHandle();
};

class CDirChangeWatcher : public CCriticalSection
{
public:
    CDirChangeWatcher();
    virtual ~CDirChangeWatcher();

    void CloseCompPort();
    void CloseThread();
    DWORD WatchProc();

    static DWORD WINAPI WatchThreadProc(LPVOID pThis)
    {
        CDirChangeWatcher *pWatcher = reinterpret_cast<CDirChangeWatcher *>(pThis);
        assert(pWatcher);
        return pWatcher->WatchProc();
    }

    void ProcessChangeNotifications(CDirWatchItem *pWatchItem);

    BOOL IsWatchingDir(LPCWSTR pszDir);
    DWORD WatchDir(LPCWSTR pszDir, DWORD dwChangeFilter,
                   CDirChangeHandler *pChangeHandler, BOOL bWatchSubDir);

    INT FindWatchItem(LPCWSTR pszDir) const;
    INT FindWatchItem(CDirWatchItem *pWatchItem) const;
    INT FindWatchItem(CDirChangeHandler *pChangeHandler) const;
    INT AddWatchItem(CDirWatchItem *pWatchItem);

    BOOL UnwatchDir(LPCWSTR pszDir);
    BOOL UnwatchDir(CDirWatchItem *pWatchItem);
    BOOL UnwatchDir(CDirChangeHandler *pChangeHandler);
    BOOL UnwatchAll();

protected:
    HANDLE m_hCompPort;
    HANDLE m_hThread;
    std::vector<CDirWatchItem*> m_WatchItems;
};
